"""Add index to entities

Revision ID: 6ccf1ef385a6
Revises: 36cdd078f890
Create Date: 2024-12-23 19:26:13.287735

"""
from typing import Sequence, Union

from alembic import op


# revision identifiers, used by Alembic.
revision: str = '6ccf1ef385a6'
down_revision: Union[str, None] = '36cdd078f890'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('entities_class_group', 'entities', ['class', 'groupname'], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('entities_class_group', table_name='entities')

    # ### end Alembic commands ###
